@@ -1,9 +1,9 @@
 /*
- *	$Id: lspci.c,v 1.16 1998/10/19 13:36:19 mj Exp $
+ *	$Id: lspci.c,v 1.35 2000/01/23 05:57:04 mj Exp $
  *
  *	Linux PCI Utilities -- List All PCI Devices
  *
- *	Copyright (c) 1997, 1998 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
+ *	Copyright (c) 1997--1999 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
  *
  *	Can be freely distributed and used under the terms of the GNU GPL.
  */
@@ -11,7 +11,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
-#include <fcntl.h>
+#include <stdarg.h>
 #include <unistd.h>
 
 #include "pciutils.h"
@@ -24,9 +24,9 @@ static int show_hex;			/* Show contents of config space as hexadecimal numbers *
 static struct pci_filter filter;	/* Device filter */
 static int show_tree;			/* Show bus tree */
 static int machine_readable;		/* Generate machine-readable output */
-static char *pci_dir = PROC_BUS_PCI;
+static int map_mode;			/* Bus mapping mode enabled */
 
-static char options[] = "nvbxs:d:ti:p:m";
+static char options[] = "nvbxs:d:ti:mgM" GENERIC_OPTIONS ;
 
 static char help_msg[] = "\
 Usage: lspci [<switches>]\n\
@@ -39,11 +39,16 @@ Usage: lspci [<switches>]\n\
 -d [<vendor>]:[<device>]\tShow only selected devices\n\
 -t\t\tShow bus tree\n\
 -m\t\tProduce machine-readable output\n\
--i <file>\tUse specified ID database instead of " ETC_PCI_IDS "\n\
--p <dir>\tUse specified bus directory instead of " PROC_BUS_PCI "\n\
-";
+-i <file>\tUse specified ID database instead of %s\n\
+-M\t\tEnable `bus mapping' mode (dangerous; root only)\n"
+GENERIC_HELP
+;
 
-/* Format strings used for IRQ numbers */
+/* Communication with libpci */
+
+static struct pci_access *pacc;
+
+/* Format strings used for IRQ numbers and memory addresses */
 
 #ifdef ARCH_SPARC64
 #define IRQ_FORMAT "%08x"
@@ -51,122 +56,94 @@ Usage: lspci [<switches>]\n\
 #define IRQ_FORMAT "%d"
 #endif
 
+#ifdef HAVE_64BIT_ADDRESS
+#ifdef HAVE_LONG_ADDRESS
+#define ADDR_FORMAT "%016Lx"
+#else
+#define ADDR_FORMAT "%016lx"
+#endif
+#else
+#define ADDR_FORMAT "%08lx"
+#endif
+
+#ifdef ARCH_SPARC64
+#define IO_FORMAT "%016Lx"
+#elif defined(HAVE_LONG_ADDRESS)
+#define IO_FORMAT "%04Lx"
+#else
+#define IO_FORMAT "%04lx"
+#endif
+
 /* Our view of the PCI bus */
 
 struct device {
   struct device *next;
-  byte bus, devfn;
-  word vendid, devid;
-  unsigned int kernel_irq;
-  unsigned long kernel_base_addr[6], kernel_rom_base_addr;
+  struct pci_dev *dev;
+  unsigned int config_cnt;
   byte config[256];
 };
 
-static struct device *first_dev, **last_dev = &first_dev;
+static struct device *first_dev;
 
-/* Miscellaneous routines */
-
-void *
-xmalloc(unsigned int howmuch)
+static struct device *
+scan_device(struct pci_dev *p)
 {
-  void *p = malloc(howmuch);
-  if (!p)
+  int how_much = (show_hex > 2) ? 256 : 64;
+  struct device *d;
+
+  if (!pci_filter_match(&filter, p))
+    return NULL;
+  d = xmalloc(sizeof(struct device));
+  bzero(d, sizeof(*d));
+  d->dev = p;
+  if (!pci_read_block(p, 0, d->config, how_much))
+    die("Unable to read %d bytes of configuration space.", how_much);
+  if (how_much < 128 && (d->config[PCI_HEADER_TYPE] & 0x7f) == PCI_HEADER_TYPE_CARDBUS)
     {
-      fprintf(stderr, "lspci: Unable to allocate %d bytes of memory\n", howmuch);
-      exit(1);
+      /* For cardbus bridges, we need to fetch 64 bytes more to get the full standard header... */
+      if (!pci_read_block(p, 64, d->config+64, 64))
+	die("Unable to read cardbus bridge extension data.");
+      how_much = 128;
     }
-  return p;
-}
-
-/* Interface for /proc/bus/pci */
-
-static void
-scan_dev_list(void)
-{
-  FILE *f;
-  byte line[256];
-  byte name[256];
-
-  sprintf(name, "%s/devices", pci_dir);
-  if (! (f = fopen(name, "r")))
-    {
-      perror(name);
-      exit(1);
-    }
-  while (fgets(line, sizeof(line), f))
-    {
-      struct device *d = xmalloc(sizeof(struct device));
-      unsigned int dfn, vend;
-
-      bzero(d, sizeof(*d));
-      sscanf(line, "%x %x %x %lx %lx %lx %lx %lx %lx %lx",
-	     &dfn,
-	     &vend,
-	     &d->kernel_irq,
-	     &d->kernel_base_addr[0],
-	     &d->kernel_base_addr[1],
-	     &d->kernel_base_addr[2],
-	     &d->kernel_base_addr[3],
-	     &d->kernel_base_addr[4],
-	     &d->kernel_base_addr[5],
-	     &d->kernel_rom_base_addr);
-      d->bus = dfn >> 8U;
-      d->devfn = dfn & 0xff;
-      d->vendid = vend >> 16U;
-      d->devid = vend & 0xffff;
-      if (filter_match(&filter, d->bus, d->devfn, d->vendid, d->devid))
-	{
-	  *last_dev = d;
-	  last_dev = &d->next;
-	  d->next = NULL;
-	}
-    }
-  fclose(f);
-}
-
-static inline void
-make_proc_pci_name(struct device *d, char *p)
-{
-  sprintf(p, "%s/%02x/%02x.%x",
-	  pci_dir, d->bus, PCI_SLOT(d->devfn), PCI_FUNC(d->devfn));
+  d->config_cnt = how_much;
+  pci_setup_cache(p, d->config, d->config_cnt);
+  pci_fill_info(p, PCI_FILL_IDENT | PCI_FILL_IRQ | PCI_FILL_BASES | PCI_FILL_ROM_BASE | PCI_FILL_SIZES);
+  return d;
 }
 
 static void
-scan_config(void)
+scan_devices(void)
 {
   struct device *d;
-  char name[64];
-  int fd, res;
-  int how_much = (show_hex > 2) ? 256 : 64;
+  struct pci_dev *p;
 
-  for(d=first_dev; d; d=d->next)
-    {
-      make_proc_pci_name(d, name);
-      if ((fd = open(name, O_RDONLY)) < 0)
-	{
-	  fprintf(stderr, "lspci: Unable to open %s: %m\n", name);
-	  exit(1);
-	}
-      res = read(fd, d->config, how_much);
-      if (res < 0)
-	{
-	  fprintf(stderr, "lspci: Error reading %s: %m\n", name);
-	  exit(1);
-	}
-      if (res != how_much)
-	{
-	  fprintf(stderr, "lspci: Only %d bytes of config space available to you\n", res);
-	  exit(1);
-	}
-      close(fd);
-    }
+  pci_scan_bus(pacc);
+  for(p=pacc->devices; p; p=p->next)
+    if (d = scan_device(p))
+      {
+	d->next = first_dev;
+	first_dev = d;
+      }
 }
 
-static void
-scan_proc(void)
+static int
+check_root(void)
 {
-  scan_dev_list();
-  scan_config();
+  static int is_root = -1;
+
+  if (is_root < 0)
+    is_root = !geteuid();
+  return is_root;
+}
+
+static int
+config_fetch(struct device *d, unsigned int pos, unsigned int len)
+{
+  if (pos + len < d->config_cnt)
+    return 1;
+  if (pacc->method != PCI_ACCESS_DUMP && !check_root())
+    return 0;
+  return pci_read_block(d->dev, pos, d->config + pos, len);
 }
 
 /* Config space accesses */
@@ -197,16 +174,20 @@ get_conf_long(struct device *d, unsigned int pos)
 static int
 compare_them(const void *A, const void *B)
 {
-  const struct device *a = *(const struct device **)A;
-  const struct device *b = *(const struct device **)B;
+  const struct pci_dev *a = (*(const struct device **)A)->dev;
+  const struct pci_dev *b = (*(const struct device **)B)->dev;
 
   if (a->bus < b->bus)
     return -1;
   if (a->bus > b->bus)
     return 1;
-  if (a->devfn < b->devfn)
+  if (a->dev < b->dev)
     return -1;
-  if (a->devfn > b->devfn)
+  if (a->dev > b->dev)
+    return 1;
+  if (a->func < b->func)
+    return -1;
+  if (a->func > b->func)
     return 1;
   return 0;
 }
@@ -214,7 +195,7 @@ compare_them(const void *A, const void *B)
 static void
 sort_them(void)
 {
-  struct device **index, **h;
+  struct device **index, **h, **last_dev;
   int cnt;
   struct device *d;
 
@@ -238,54 +219,95 @@ sort_them(void)
 
 /* Normal output */
 
+#define FLAG(x,y) ((x & y) ? '+' : '-')
+
 static void
 show_terse(struct device *d)
 {
   int c;
+  struct pci_dev *p = d->dev;
+  byte classbuf[128], devbuf[128];
 
   printf("%02x:%02x.%x %s: %s",
-	 d->bus,
-	 PCI_SLOT(d->devfn),
-	 PCI_FUNC(d->devfn),
-	 lookup_class(get_conf_word(d, PCI_CLASS_DEVICE)),
-	 lookup_device_full(d->vendid, d->devid));
+	 p->bus,
+	 p->dev,
+	 p->func,
+	 pci_lookup_name(pacc, classbuf, sizeof(classbuf),
+			 PCI_LOOKUP_CLASS,
+			 get_conf_word(d, PCI_CLASS_DEVICE), 0, 0, 0),
+	 pci_lookup_name(pacc, devbuf, sizeof(devbuf),
+			 PCI_LOOKUP_VENDOR | PCI_LOOKUP_DEVICE,
+			 p->vendor_id, p->device_id, 0, 0));
   if (c = get_conf_byte(d, PCI_REVISION_ID))
     printf(" (rev %02x)", c);
-  if (verbose && (c = get_conf_byte(d, PCI_CLASS_PROG)))
-    printf(" (prog-if %02x)", c);
+  if (verbose)
+    {
+      char *x;
+      c = get_conf_byte(d, PCI_CLASS_PROG);
+      x = pci_lookup_name(pacc, devbuf, sizeof(devbuf),
+			  PCI_LOOKUP_PROGIF,
+			  get_conf_word(d, PCI_CLASS_DEVICE), c, 0, 0);
+      if (c || x)
+	{
+	  printf(" (prog-if %02x", c);
+	  if (x)
+	    printf(" [%s]", x);
+	  putchar(')');
+	}
+    }
   putchar('\n');
 }
 
+static void
+show_size(pciaddr_t x)
+{
+  if (!x)
+    return;
+  printf(" [size=");
+  if (x < 1024)
+    printf("%d", (int) x);
+  else if (x < 1048576)
+    printf("%dK", (int)(x / 1024));
+  else if (x < 0x80000000)
+    printf("%dM", (int)(x / 1048576));
+  else
+    printf(ADDR_FORMAT, x);
+  putchar(']');
+}
+
 static void
 show_bases(struct device *d, int cnt)
 {
+  struct pci_dev *p = d->dev;
   word cmd = get_conf_word(d, PCI_COMMAND);
   int i;
 
   for(i=0; i<cnt; i++)
     {
-      unsigned long pos;
-      unsigned int flg = get_conf_long(d, PCI_BASE_ADDRESS_0 + 4*i);
-      if (buscentric_view)
-	pos = flg;
-      else
-	{
-	  pos = d->kernel_base_addr[i];
-	  if (!pos)
-	    continue;
-	}
-      if (pos == 0xffffffff)
+      pciaddr_t pos = p->base_addr[i];
+      pciaddr_t len = (p->known_fields & PCI_FILL_SIZES) ? p->size[i] : 0;
+      u32 flg = get_conf_long(d, PCI_BASE_ADDRESS_0 + 4*i);
+      if (flg == 0xffffffff)
+	flg = 0;
+      if (!pos && !flg && !len)
 	continue;
       if (verbose > 1)
 	printf("\tRegion %d: ", i);
       else
 	putchar('\t');
+      if (pos && !flg)			/* Reported by the OS, but not by the device */
+	{
+	  printf("[virtual] ");
+	  flg = pos;
+	}
       if (flg & PCI_BASE_ADDRESS_SPACE_IO)
 	{
-	  unsigned long a = pos & PCI_BASE_ADDRESS_IO_MASK;
+	  pciaddr_t a = pos & PCI_BASE_ADDRESS_IO_MASK;
 	  printf("I/O ports at ");
 	  if (a)
-	    printf("%04lx", a);
+	    printf(IO_FORMAT, a);
+	  else if (flg & PCI_BASE_ADDRESS_IO_MASK)
+	    printf("<ignored>");
 	  else
 	    printf("<unassigned>");
 	  if (!(cmd & PCI_COMMAND_IO))
@@ -294,31 +316,39 @@ show_bases(struct device *d, int cnt)
       else
 	{
 	  int t = flg & PCI_BASE_ADDRESS_MEM_TYPE_MASK;
-	  unsigned long a = pos & PCI_BASE_ADDRESS_MEM_MASK;
-	  int x64 = 0;
+	  pciaddr_t a = pos & PCI_ADDR_MEM_MASK;
+	  int done = 0;
+	  u32 z = 0;
+
 	  printf("Memory at ");
 	  if (t == PCI_BASE_ADDRESS_MEM_TYPE_64)
 	    {
-	      if (i < cnt - 1)
+	      if (i >= cnt - 1)
 		{
-		  u32 z;
-		  i++;
-		  z = get_conf_long(d, PCI_BASE_ADDRESS_0 + 4*i);
-		  if (buscentric_view)
-		    printf("%08x", z);
-		  if (z)
-		    x64 = 1;
+		  printf("<invalid-64bit-slot>");
+		  done = 1;
 		}
 	      else
 		{
-		  printf("????????");
-		  x64 = 1;
+		  i++;
+		  z = get_conf_long(d, PCI_BASE_ADDRESS_0 + 4*i);
+		  if (buscentric_view)
+		    {
+		      if (a || z)
+			printf("%08x" ADDR_FORMAT, z, a);
+		      else
+			printf("<unassigned>");
+		      done = 1;
+		    }
 		}
 	    }
-	  if (x64 || a)
-	    printf("%08lx", a);
-	  else
-	    printf("<unassigned>");
+	  if (!done)
+	    {
+	      if (a)
+		printf(ADDR_FORMAT, a);
+	      else
+		printf(((flg & PCI_BASE_ADDRESS_MEM_MASK) || z) ? "<ignored>" : "<unassigned>");
+	    }
 	  printf(" (%s, %sprefetchable)",
 		 (t == PCI_BASE_ADDRESS_MEM_TYPE_32) ? "32-bit" :
 		 (t == PCI_BASE_ADDRESS_MEM_TYPE_64) ? "64-bit" :
@@ -327,20 +357,202 @@ show_bases(struct device *d, int cnt)
 	  if (!(cmd & PCI_COMMAND_MEMORY))
 	    printf(" [disabled]");
 	}
+      show_size(len);
       putchar('\n');
     }
 }
 
+static void
+show_pm(struct device *d, int where, int cap)
+{
+  int t;
+
+  printf("Power Management version %d\n", cap & PCI_PM_CAP_VER_MASK);
+  if (verbose < 2)
+    return;
+  printf("\t\tFlags: PMEClk%c AuxPwr%c DSI%c D1%c D2%c PME%c\n",
+	 FLAG(cap, PCI_PM_CAP_PME_CLOCK),
+	 FLAG(cap, PCI_PM_CAP_AUX_POWER),
+	 FLAG(cap, PCI_PM_CAP_DSI),
+	 FLAG(cap, PCI_PM_CAP_D1),
+	 FLAG(cap, PCI_PM_CAP_D2),
+	 FLAG(cap, PCI_PM_CAP_PME));
+  config_fetch(d, where + PCI_PM_CTRL, PCI_PM_SIZEOF - PCI_PM_CTRL);
+  t = get_conf_word(d, where + PCI_PM_CTRL);
+  printf("\t\tStatus: D%d PME-Enable%c DSel=%x DScale=%x PME%c\n",
+	 t & PCI_PM_CTRL_STATE_MASK,
+	 FLAG(t, PCI_PM_CTRL_PME_ENABLE),
+	 (t & PCI_PM_CTRL_DATA_SEL_MASK) >> 9,
+	 (t & PCI_PM_CTRL_DATA_SCALE_MASK) >> 13,
+	 FLAG(t, PCI_PM_CTRL_PME_STATUS));
+}
+
+static void
+format_agp_rate(int rate, char *buf)
+{
+  char *c = buf;
+  int i;
+
+  for(i=0; i<2; i++)
+    if (rate & (1 << i))
+      {
+	if (c != buf)
+	  *c++ = ',';
+	*c++ = 'x';
+	*c++ = '0' + (1 << i);
+      }
+  if (c != buf)
+    *c = 0;
+  else
+    strcpy(buf, "<none>");
+}
+
+static void
+show_agp(struct device *d, int where, int cap)
+{
+  u32 t;
+  char rate[8];
+
+  t = cap & 0xff;
+  printf("AGP version %x.%x\n", cap/16, cap%16);
+  if (verbose < 2)
+    return;
+  config_fetch(d, where + PCI_AGP_STATUS, PCI_AGP_SIZEOF - PCI_AGP_STATUS);
+  t = get_conf_long(d, where + PCI_AGP_STATUS);
+  format_agp_rate(t & 7, rate);
+  printf("\t\tStatus: RQ=%d SBA%c 64bit%c FW%c Rate=%s\n",
+	 (t & PCI_AGP_STATUS_RQ_MASK) >> 24U,
+	 FLAG(t, PCI_AGP_STATUS_SBA),
+	 FLAG(t, PCI_AGP_STATUS_64BIT),
+	 FLAG(t, PCI_AGP_STATUS_FW),
+	 rate);
+  t = get_conf_long(d, where + PCI_AGP_COMMAND);
+  format_agp_rate(t & 7, rate);
+  printf("\t\tCommand: RQ=%d SBA%c AGP%c 64bit%c FW%c Rate=%s\n",
+	 (t & PCI_AGP_COMMAND_RQ_MASK) >> 24U,
+	 FLAG(t, PCI_AGP_COMMAND_SBA),
+	 FLAG(t, PCI_AGP_COMMAND_AGP),
+	 FLAG(t, PCI_AGP_COMMAND_64BIT),
+	 FLAG(t, PCI_AGP_COMMAND_FW),
+	 rate);
+}
+
+static void
+show_rom(struct device *d)
+{
+  struct pci_dev *p = d->dev;
+  pciaddr_t rom = p->rom_base_addr;
+  pciaddr_t len = (p->known_fields & PCI_FILL_SIZES) ? p->rom_size : 0;
+
+  if (!rom && !len)
+    return;
+  printf("\tExpansion ROM at ");
+  if (rom & PCI_ROM_ADDRESS_MASK)
+    printf(ADDR_FORMAT, rom & PCI_ROM_ADDRESS_MASK);
+  else
+    printf("<unassigned>");
+  if (!(rom & PCI_ROM_ADDRESS_ENABLE))
+    printf(" [disabled]");
+  show_size(len);
+  putchar('\n');
+}
+
+static void
+show_msi(struct device *d, int where, int cap)
+{
+  int is64;
+  u32 t;
+  u16 w;
+
+  printf("Message Signalled Interrupts: 64bit%c Queue=%d/%d Enable%c\n",
+	 FLAG(cap, PCI_MSI_FLAGS_64BIT),
+	 (cap & PCI_MSI_FLAGS_QSIZE) >> 4,
+	 (cap & PCI_MSI_FLAGS_QMASK) >> 1,
+	 FLAG(cap, PCI_MSI_FLAGS_ENABLE));
+  if (verbose < 2)
+    return;
+  is64 = cap & PCI_MSI_FLAGS_64BIT;
+  config_fetch(d, where + PCI_MSI_ADDRESS_LO, (is64 ? PCI_MSI_DATA_64 : PCI_MSI_DATA_32) + 2 - PCI_MSI_ADDRESS_LO);
+  printf("\t\tAddress: ");
+  if (is64)
+    {
+      t = get_conf_long(d, where + PCI_MSI_ADDRESS_HI);
+      w = get_conf_word(d, where + PCI_MSI_DATA_64);
+      printf("%08x", t);
+    }
+  else
+    w = get_conf_word(d, where + PCI_MSI_DATA_32);
+  t = get_conf_long(d, where + PCI_MSI_ADDRESS_LO);
+  printf("%08x  Data: %04x\n", t, w);
+}
+
+static void
+show_slotid(int cap)
+{
+  int esr = cap & 0xff;
+  int chs = cap >> 8;
+
+  printf("Slot ID: %d slots, First%c, chassis %02x\n",
+	 esr & PCI_SID_ESR_NSLOTS,
+	 FLAG(esr, PCI_SID_ESR_FIC),
+	 chs);
+}
+
+static void
+show_caps(struct device *d)
+{
+  if (get_conf_word(d, PCI_STATUS) & PCI_STATUS_CAP_LIST)
+    {
+      int where = get_conf_byte(d, PCI_CAPABILITY_LIST) & ~3;
+      while (where)
+	{
+	  int id, next, cap;
+	  printf("\tCapabilities: ");
+	  if (!config_fetch(d, where, 4))
+	    {
+	      puts("<available only to root>");
+	      break;
+	    }
+	  id = get_conf_byte(d, where + PCI_CAP_LIST_ID);
+	  next = get_conf_byte(d, where + PCI_CAP_LIST_NEXT) & ~3;
+	  cap = get_conf_word(d, where + PCI_CAP_FLAGS);
+	  printf("[%02x] ", where);
+	  if (id == 0xff)
+	    {
+	      printf("<chain broken>\n");
+	      break;
+	    }
+	  switch (id)
+	    {
+	    case PCI_CAP_ID_PM:
+	      show_pm(d, where, cap);
+	      break;
+	    case PCI_CAP_ID_AGP:
+	      show_agp(d, where, cap);
+	      break;
+	    case PCI_CAP_ID_VPD:
+	      printf("Vital Product Data\n");
+	      break;
+	    case PCI_CAP_ID_SLOTID:
+	      show_slotid(cap);
+	      break;
+	    case PCI_CAP_ID_MSI:
+	      show_msi(d, where, cap);
+	      break;
+	    default:
+	      printf("#%02x [%04x]\n", id, cap);
+	    }
+	  where = next;
+	}
+    }
+}
+
 static void
 show_htype0(struct device *d)
 {
-  unsigned long rom = buscentric_view ? get_conf_long(d, PCI_ROM_ADDRESS) : d->kernel_rom_base_addr;
-
   show_bases(d, 6);
-
-  if (rom & 1)
-    printf("\tExpansion ROM at %08lx%s\n", rom & PCI_ROM_ADDRESS_MASK,
-	   (rom & PCI_ROM_ADDRESS_ENABLE) ? "" : " [disabled]");
+  show_rom(d);
+  show_caps(d);
 }
 
 static void
@@ -355,8 +567,8 @@ show_htype1(struct device *d)
   u32 pref_base = get_conf_word(d, PCI_PREF_MEMORY_BASE);
   u32 pref_limit = get_conf_word(d, PCI_PREF_MEMORY_LIMIT);
   u32 pref_type = pref_base & PCI_PREF_RANGE_TYPE_MASK;
-  unsigned long rom = buscentric_view ? get_conf_long(d, PCI_ROM_ADDRESS) : d->kernel_rom_base_addr;
   word brc = get_conf_word(d, PCI_BRIDGE_CONTROL);
+  int verb = verbose > 2;
 
   show_bases(d, 2);
   printf("\tBus: primary=%02x, secondary=%02x, subordinate=%02x, sec-latency=%d\n",
@@ -377,53 +589,57 @@ show_htype1(struct device *d)
 	  io_base |= (get_conf_word(d, PCI_IO_BASE_UPPER16) << 16);
 	  io_limit |= (get_conf_word(d, PCI_IO_LIMIT_UPPER16) << 16);
 	}
-      if (io_base)
+      if (io_base <= io_limit || verb)
 	printf("\tI/O behind bridge: %08x-%08x\n", io_base, io_limit+0xfff);
     }
 
   if (mem_type != (mem_limit & PCI_MEMORY_RANGE_TYPE_MASK) ||
       mem_type)
     printf("\t!!! Unknown memory range types %x/%x\n", mem_base, mem_limit);
-  else if (mem_base)
+  else
     {
       mem_base = (mem_base & PCI_MEMORY_RANGE_MASK) << 16;
       mem_limit = (mem_limit & PCI_MEMORY_RANGE_MASK) << 16;
-      printf("\tMemory behind bridge: %08x-%08x\n", mem_base, mem_limit + 0xfffff);
+      if (mem_base <= mem_limit || verb)
+	printf("\tMemory behind bridge: %08x-%08x\n", mem_base, mem_limit + 0xfffff);
     }
 
   if (pref_type != (pref_limit & PCI_PREF_RANGE_TYPE_MASK) ||
       (pref_type != PCI_PREF_RANGE_TYPE_32 && pref_type != PCI_PREF_RANGE_TYPE_64))
     printf("\t!!! Unknown prefetchable memory range types %x/%x\n", pref_base, pref_limit);
-  else if (pref_base)
+  else
     {
       pref_base = (pref_base & PCI_PREF_RANGE_MASK) << 16;
       pref_limit = (pref_limit & PCI_PREF_RANGE_MASK) << 16;
-      if (pref_type == PCI_PREF_RANGE_TYPE_32)
-	printf("\tPrefetchable memory behind bridge: %08x-%08x\n", pref_base, pref_limit);
-      else
-	printf("\tPrefetchable memory behind bridge: %08x%08x-%08x%08x\n",
-	       get_conf_long(d, PCI_PREF_BASE_UPPER32),
-	       pref_base,
-	       get_conf_long(d, PCI_PREF_LIMIT_UPPER32),
-	       pref_limit);
+      if (pref_base <= pref_limit || verb)
+	{
+	  if (pref_type == PCI_PREF_RANGE_TYPE_32)
+	    printf("\tPrefetchable memory behind bridge: %08x-%08x\n", pref_base, pref_limit + 0xfffff);
+	  else
+	    printf("\tPrefetchable memory behind bridge: %08x%08x-%08x%08x\n",
+		   get_conf_long(d, PCI_PREF_BASE_UPPER32),
+		   pref_base,
+		   get_conf_long(d, PCI_PREF_LIMIT_UPPER32),
+		   pref_limit);
+	}
     }
 
   if (get_conf_word(d, PCI_SEC_STATUS) & PCI_STATUS_SIG_SYSTEM_ERROR)
     printf("\tSecondary status: SERR\n");
 
-  if (rom & 1)
-    printf("\tExpansion ROM at %08lx%s\n", rom & PCI_ROM_ADDRESS_MASK,
-	   (rom & PCI_ROM_ADDRESS_ENABLE) ? "" : " [disabled]");
+  show_rom(d);
 
   if (verbose > 1)
     printf("\tBridgeCtl: Parity%c SERR%c NoISA%c VGA%c MAbort%c >Reset%c FastB2B%c\n",
-	   (brc & PCI_BRIDGE_CTL_PARITY) ? '+' : '-',
-	   (brc & PCI_BRIDGE_CTL_SERR) ? '+' : '-',
-	   (brc & PCI_BRIDGE_CTL_NO_ISA) ? '+' : '-',
-	   (brc & PCI_BRIDGE_CTL_VGA) ? '+' : '-',
-	   (brc & PCI_BRIDGE_CTL_MASTER_ABORT) ? '+' : '-',
-	   (brc & PCI_BRIDGE_CTL_BUS_RESET) ? '+' : '-',
-	   (brc & PCI_BRIDGE_CTL_FAST_BACK) ? '+' : '-');
+	   FLAG(brc, PCI_BRIDGE_CTL_PARITY),
+	   FLAG(brc, PCI_BRIDGE_CTL_SERR),
+	   FLAG(brc, PCI_BRIDGE_CTL_NO_ISA),
+	   FLAG(brc, PCI_BRIDGE_CTL_VGA),
+	   FLAG(brc, PCI_BRIDGE_CTL_MASTER_ABORT),
+	   FLAG(brc, PCI_BRIDGE_CTL_BUS_RESET),
+	   FLAG(brc, PCI_BRIDGE_CTL_FAST_BACK));
+
+  show_caps(d);
 }
 
 static void
@@ -433,6 +649,7 @@ show_htype2(struct device *d)
   word cmd = get_conf_word(d, PCI_COMMAND);
   word brc = get_conf_word(d, PCI_CB_BRIDGE_CONTROL);
   word exca = get_conf_word(d, PCI_CB_LEGACY_MODE_BASE);
+  int verb = verbose > 2;
 
   show_bases(d, 1);
   printf("\tBus: primary=%02x, secondary=%02x, subordinate=%02x, sec-latency=%d\n",
@@ -445,8 +662,8 @@ show_htype2(struct device *d)
       int p = 8*i;
       u32 base = get_conf_long(d, PCI_CB_MEMORY_BASE_0 + p);
       u32 limit = get_conf_long(d, PCI_CB_MEMORY_LIMIT_0 + p);
-      if (limit > base)
-	printf("Memory window %d: %08x-%08x%s%s\n", i, base, limit,
+      if (limit > base || verb)
+	printf("\tMemory window %d: %08x-%08x%s%s\n", i, base, limit,
 	       (cmd & PCI_COMMAND_MEMORY) ? "" : " [disabled]",
 	       (brc & (PCI_CB_BRIDGE_CTL_PREFETCH_MEM0 << i)) ? " (prefetchable)" : "");
     }
@@ -461,25 +678,24 @@ show_htype2(struct device *d)
 	  limit &= 0xffff;
 	}
       base &= PCI_CB_IO_RANGE_MASK;
-      if (!base)
-	continue;
       limit = (limit & PCI_CB_IO_RANGE_MASK) + 3;
-      printf("I/O window %d: %08x-%08x%s\n", i, base, limit,
-	     (cmd & PCI_COMMAND_IO) ? "" : " [disabled]");
+      if (base <= limit || verb)
+	printf("\tI/O window %d: %08x-%08x%s\n", i, base, limit,
+	       (cmd & PCI_COMMAND_IO) ? "" : " [disabled]");
     }
 
   if (get_conf_word(d, PCI_CB_SEC_STATUS) & PCI_STATUS_SIG_SYSTEM_ERROR)
     printf("\tSecondary status: SERR\n");
   if (verbose > 1)
     printf("\tBridgeCtl: Parity%c SERR%c ISA%c VGA%c MAbort%c >Reset%c 16bInt%c PostWrite%c\n",
-	   (brc & PCI_CB_BRIDGE_CTL_PARITY) ? '+' : '-',
-	   (brc & PCI_CB_BRIDGE_CTL_SERR) ? '+' : '-',
-	   (brc & PCI_CB_BRIDGE_CTL_ISA) ? '+' : '-',
-	   (brc & PCI_CB_BRIDGE_CTL_VGA) ? '+' : '-',
-	   (brc & PCI_CB_BRIDGE_CTL_MASTER_ABORT) ? '+' : '-',
-	   (brc & PCI_CB_BRIDGE_CTL_CB_RESET) ? '+' : '-',
-	   (brc & PCI_CB_BRIDGE_CTL_16BIT_INT) ? '+' : '-',
-	   (brc & PCI_CB_BRIDGE_CTL_POST_WRITES) ? '+' : '-');
+	   FLAG(brc, PCI_CB_BRIDGE_CTL_PARITY),
+	   FLAG(brc, PCI_CB_BRIDGE_CTL_SERR),
+	   FLAG(brc, PCI_CB_BRIDGE_CTL_ISA),
+	   FLAG(brc, PCI_CB_BRIDGE_CTL_VGA),
+	   FLAG(brc, PCI_CB_BRIDGE_CTL_MASTER_ABORT),
+	   FLAG(brc, PCI_CB_BRIDGE_CTL_CB_RESET),
+	   FLAG(brc, PCI_CB_BRIDGE_CTL_16BIT_INT),
+	   FLAG(brc, PCI_CB_BRIDGE_CTL_POST_WRITES));
   if (exca)
     printf("\t16-bit legacy interface ports at %04x\n", exca);
 }
@@ -487,6 +703,7 @@ show_htype2(struct device *d)
 static void
 show_verbose(struct device *d)
 {
+  struct pci_dev *p = d->dev;
   word status = get_conf_word(d, PCI_STATUS);
   word cmd = get_conf_word(d, PCI_COMMAND);
   word class = get_conf_word(d, PCI_CLASS_DEVICE);
@@ -496,9 +713,9 @@ show_verbose(struct device *d)
   byte cache_line = get_conf_byte(d, PCI_CACHE_LINE_SIZE);
   byte max_lat, min_gnt;
   byte int_pin = get_conf_byte(d, PCI_INTERRUPT_PIN);
-  byte int_line = get_conf_byte(d, PCI_INTERRUPT_LINE);
-  unsigned int irq;
+  unsigned int irq = p->irq;
   word subsys_v, subsys_d;
+  char ssnamebuf[256];
 
   show_terse(d);
 
@@ -519,7 +736,7 @@ show_verbose(struct device *d)
     case PCI_HEADER_TYPE_BRIDGE:
       if (class != PCI_CLASS_BRIDGE_PCI)
 	goto badhdr;
-      irq = int_line = int_pin = min_gnt = max_lat = 0;
+      irq = int_pin = min_gnt = max_lat = 0;
       subsys_v = subsys_d = 0;
       break;
     case PCI_HEADER_TYPE_CARDBUS:
@@ -534,40 +751,39 @@ show_verbose(struct device *d)
       return;
     }
 
-  if (buscentric_view)
-    irq = int_line;
-  else
-    irq = d->kernel_irq;
-
-  if (verbose && subsys_v && subsys_v != 0xffff)
-    printf("\tSubsystem: %s\n", lookup_subsys_device_full(subsys_v, subsys_d));
+  if (subsys_v && subsys_v != 0xffff)
+    printf("\tSubsystem: %s\n",
+	   pci_lookup_name(pacc, ssnamebuf, sizeof(ssnamebuf),
+			   PCI_LOOKUP_SUBSYSTEM | PCI_LOOKUP_VENDOR | PCI_LOOKUP_DEVICE,
+			   p->vendor_id, p->device_id, subsys_v, subsys_d));
 
   if (verbose > 1)
     {
       printf("\tControl: I/O%c Mem%c BusMaster%c SpecCycle%c MemWINV%c VGASnoop%c ParErr%c Stepping%c SERR%c FastB2B%c\n",
-	     (cmd & PCI_COMMAND_IO) ? '+' : '-',
-	     (cmd & PCI_COMMAND_MEMORY) ? '+' : '-',
-	     (cmd & PCI_COMMAND_MASTER) ? '+' : '-',
-	     (cmd & PCI_COMMAND_SPECIAL) ? '+' : '-',
-	     (cmd & PCI_COMMAND_INVALIDATE) ? '+' : '-',
-	     (cmd & PCI_COMMAND_VGA_PALETTE) ? '+' : '-',
-	     (cmd & PCI_COMMAND_PARITY) ? '+' : '-',
-	     (cmd & PCI_COMMAND_WAIT) ? '+' : '-',
-	     (cmd & PCI_COMMAND_SERR) ? '+' : '-',
-	     (cmd & PCI_COMMAND_FAST_BACK) ? '+' : '-');
-      printf("\tStatus: 66Mhz%c UDF%c FastB2B%c ParErr%c DEVSEL=%s >TAbort%c <TAbort%c <MAbort%c >SERR%c <PERR%c\n",
-	     (status & PCI_STATUS_66MHZ) ? '+' : '-',
-	     (status & PCI_STATUS_UDF) ? '+' : '-',
-	     (status & PCI_STATUS_FAST_BACK) ? '+' : '-',
-	     (status & PCI_STATUS_PARITY) ? '+' : '-',
+	     FLAG(cmd, PCI_COMMAND_IO),
+	     FLAG(cmd, PCI_COMMAND_MEMORY),
+	     FLAG(cmd, PCI_COMMAND_MASTER),
+	     FLAG(cmd, PCI_COMMAND_SPECIAL),
+	     FLAG(cmd, PCI_COMMAND_INVALIDATE),
+	     FLAG(cmd, PCI_COMMAND_VGA_PALETTE),
+	     FLAG(cmd, PCI_COMMAND_PARITY),
+	     FLAG(cmd, PCI_COMMAND_WAIT),
+	     FLAG(cmd, PCI_COMMAND_SERR),
+	     FLAG(cmd, PCI_COMMAND_FAST_BACK));
+      printf("\tStatus: Cap%c 66Mhz%c UDF%c FastB2B%c ParErr%c DEVSEL=%s >TAbort%c <TAbort%c <MAbort%c >SERR%c <PERR%c\n",
+	     FLAG(status, PCI_STATUS_CAP_LIST),
+	     FLAG(status, PCI_STATUS_66MHZ),
+	     FLAG(status, PCI_STATUS_UDF),
+	     FLAG(status, PCI_STATUS_FAST_BACK),
+	     FLAG(status, PCI_STATUS_PARITY),
 	     ((status & PCI_STATUS_DEVSEL_MASK) == PCI_STATUS_DEVSEL_SLOW) ? "slow" :
 	     ((status & PCI_STATUS_DEVSEL_MASK) == PCI_STATUS_DEVSEL_MEDIUM) ? "medium" :
 	     ((status & PCI_STATUS_DEVSEL_MASK) == PCI_STATUS_DEVSEL_FAST) ? "fast" : "??",
-	     (status & PCI_STATUS_SIG_TARGET_ABORT) ? '+' : '-',
-	     (status & PCI_STATUS_REC_TARGET_ABORT) ? '+' : '-',
-	     (status & PCI_STATUS_REC_MASTER_ABORT) ? '+' : '-',
-	     (status & PCI_STATUS_SIG_SYSTEM_ERROR) ? '+' : '-',
-	     (status & PCI_STATUS_DETECTED_PARITY) ? '+' : '-');
+	     FLAG(status, PCI_STATUS_SIG_TARGET_ABORT),
+	     FLAG(status, PCI_STATUS_REC_TARGET_ABORT),
+	     FLAG(status, PCI_STATUS_REC_MASTER_ABORT),
+	     FLAG(status, PCI_STATUS_SIG_SYSTEM_ERROR),
+	     FLAG(status, PCI_STATUS_DETECTED_PARITY));
       if (cmd & PCI_COMMAND_MASTER)
 	{
 	  printf("\tLatency: ");
@@ -580,8 +796,9 @@ show_verbose(struct device *d)
 	    printf(", cache line size %02x", cache_line);
 	  putchar('\n');
 	}
-      if (int_pin)
-	printf("\tInterrupt: pin %c routed to IRQ " IRQ_FORMAT "\n", 'A' + int_pin - 1, irq);
+      if (int_pin || irq)
+	printf("\tInterrupt: pin %c routed to IRQ " IRQ_FORMAT "\n",
+	       (int_pin ? 'A' + int_pin - 1 : '?'), irq);
     }
   else
     {
@@ -604,11 +821,8 @@ show_verbose(struct device *d)
 	     ((status & PCI_STATUS_DEVSEL_MASK) == PCI_STATUS_DEVSEL_FAST) ? "fast" : "??");
       if (cmd & PCI_COMMAND_MASTER)
 	printf(", latency %d", latency);
-      if (int_pin)
-	if (d->kernel_irq)
-	  printf(", IRQ " IRQ_FORMAT, irq);
-	else
-	  printf(", IRQ ?");
+      if (irq)
+	printf(", IRQ " IRQ_FORMAT, irq);
       putchar('\n');
     }
 
@@ -637,10 +851,9 @@ show_verbose(struct device *d)
 static void
 show_hex_dump(struct device *d)
 {
-  int i;
-  int limit = (show_hex > 2) ? 256 : 64;
+  unsigned int i;
 
-  for(i=0; i<limit; i++)
+  for(i=0; i<d->config_cnt; i++)
     {
       if (! (i & 15))
 	printf("%02x:", i);
@@ -653,8 +866,10 @@ show_hex_dump(struct device *d)
 static void
 show_machine(struct device *d)
 {
+  struct pci_dev *p = d->dev;
   int c;
   word sv_id=0, sd_id=0;
+  char classbuf[128], vendbuf[128], devbuf[128], svbuf[128], sdbuf[128];
 
   switch (get_conf_byte(d, PCI_HEADER_TYPE) & 0x7f)
     {
@@ -670,14 +885,19 @@ show_machine(struct device *d)
 
   if (verbose)
     {
-      printf("Device:\t%02x:%02x.%x\n", d->bus, PCI_SLOT(d->devfn), PCI_FUNC(d->devfn));
-      printf("Class:\t%s\n", lookup_class(get_conf_word(d, PCI_CLASS_DEVICE)));
-      printf("Vendor:\t%s\n", lookup_vendor(d->vendid));
-      printf("Device:\t%s\n", lookup_device(d->vendid, d->devid));
+      printf("Device:\t%02x:%02x.%x\n", p->bus, p->dev, p->func);
+      printf("Class:\t%s\n",
+	     pci_lookup_name(pacc, classbuf, sizeof(classbuf), PCI_LOOKUP_CLASS, get_conf_word(d, PCI_CLASS_DEVICE), 0, 0, 0));
+      printf("Vendor:\t%s\n",
+	     pci_lookup_name(pacc, vendbuf, sizeof(vendbuf), PCI_LOOKUP_VENDOR, p->vendor_id, p->device_id, 0, 0));
+      printf("Device:\t%s\n",
+	     pci_lookup_name(pacc, devbuf, sizeof(devbuf), PCI_LOOKUP_DEVICE, p->vendor_id, p->device_id, 0, 0));
       if (sv_id && sv_id != 0xffff)
 	{
-	  printf("SVendor:\t%s\n", lookup_subsys_vendor(sv_id));
-	  printf("SDevice:\t%s\n", lookup_subsys_device(sv_id, sd_id));
+	  printf("SVendor:\t%s\n",
+		 pci_lookup_name(pacc, svbuf, sizeof(svbuf), PCI_LOOKUP_SUBSYSTEM | PCI_LOOKUP_VENDOR, p->vendor_id, p->device_id, sv_id, sd_id));
+	  printf("SDevice:\t%s\n",
+		 pci_lookup_name(pacc, sdbuf, sizeof(sdbuf), PCI_LOOKUP_SUBSYSTEM | PCI_LOOKUP_DEVICE, p->vendor_id, p->device_id, sv_id, sd_id));
 	}
       if (c = get_conf_byte(d, PCI_REVISION_ID))
 	printf("Rev:\t%02x\n", c);
@@ -686,41 +906,50 @@ show_machine(struct device *d)
     }
   else
     {
-      printf("%02x:%02x.%x ", d->bus, PCI_SLOT(d->devfn), PCI_FUNC(d->devfn));
+      printf("%02x:%02x.%x ", p->bus, p->dev, p->func);
       printf("\"%s\" \"%s\" \"%s\"",
-	     lookup_class(get_conf_word(d, PCI_CLASS_DEVICE)),
-	     lookup_vendor(d->vendid),
-	     lookup_device(d->vendid, d->devid));
+	     pci_lookup_name(pacc, classbuf, sizeof(classbuf), PCI_LOOKUP_CLASS,
+			     get_conf_word(d, PCI_CLASS_DEVICE), 0, 0, 0),
+	     pci_lookup_name(pacc, vendbuf, sizeof(vendbuf), PCI_LOOKUP_VENDOR,
+			     p->vendor_id, p->device_id, 0, 0),
+	     pci_lookup_name(pacc, devbuf, sizeof(devbuf), PCI_LOOKUP_DEVICE,
+			     p->vendor_id, p->device_id, 0, 0));
       if (c = get_conf_byte(d, PCI_REVISION_ID))
 	printf(" -r%02x", c);
       if (c = get_conf_byte(d, PCI_CLASS_PROG))
 	printf(" -p%02x", c);
       if (sv_id && sv_id != 0xffff)
-	printf(" \"%s\" \"%s\"", lookup_subsys_vendor(sv_id), lookup_subsys_device(sv_id, sd_id));
+	printf(" \"%s\" \"%s\"",
+	       pci_lookup_name(pacc, svbuf, sizeof(svbuf), PCI_LOOKUP_SUBSYSTEM | PCI_LOOKUP_VENDOR, p->vendor_id, p->device_id, sv_id, sd_id),
+	       pci_lookup_name(pacc, sdbuf, sizeof(sdbuf), PCI_LOOKUP_SUBSYSTEM | PCI_LOOKUP_DEVICE, p->vendor_id, p->device_id, sv_id, sd_id));
       else
 	printf(" \"\" \"\"");
       putchar('\n');
     }
 }
 
+static void
+show_device(struct device *d)
+{
+  if (machine_readable)
+    show_machine(d);
+  else if (verbose)
+    show_verbose(d);
+  else
+    show_terse(d);
+  if (show_hex)
+    show_hex_dump(d);
+  if (verbose || show_hex)
+    putchar('\n');
+}
+
 static void
 show(void)
 {
   struct device *d;
 
   for(d=first_dev; d; d=d->next)
-    {
-      if (machine_readable)
-	show_machine(d);
-      else if (verbose)
-	show_verbose(d);
-      else
-	show_terse(d);
-      if (show_hex)
-	show_hex_dump(d);
-      if (verbose || show_hex)
-	putchar('\n');
-    }
+    show_device(d);
 }
 
 /* Tree output */
@@ -769,15 +998,16 @@ new_bus(struct bridge *b, unsigned int n)
 static void
 insert_dev(struct device *d, struct bridge *b)
 {
+  struct pci_dev *p = d->dev;
   struct bus *bus;
 
-  if (! (bus = find_bus(b, d->bus)))
+  if (! (bus = find_bus(b, p->bus)))
     {
       struct bridge *c;
       for(c=b->child; c; c=c->next)
-	if (c->secondary <= d->bus && d->bus <= c->subordinate)
+	if (c->secondary <= p->bus && p->bus <= c->subordinate)
 	  return insert_dev(d, c);
-      bus = new_bus(b, d->bus);
+      bus = new_bus(b, p->bus);
     }
   /* Simple insertion at the end _does_ guarantee the correct order as the
    * original device list was sorted by (bus, devfn) lexicographically
@@ -877,9 +1107,11 @@ static void show_tree_bridge(struct bridge *, byte *, byte *);
 static void
 show_tree_dev(struct device *d, byte *line, byte *p)
 {
+  struct pci_dev *q = d->dev;
   struct bridge *b;
+  char namebuf[256];
 
-  p += sprintf(p, "%02x.%x", PCI_SLOT(d->devfn), PCI_FUNC(d->devfn));
+  p += sprintf(p, "%02x.%x", q->dev, q->func);
   for(b=&host_bridge; b; b=b->chain)
     if (b->br_dev == d)
       {
@@ -891,7 +1123,10 @@ show_tree_dev(struct device *d, byte *line, byte *p)
         return;
       }
   if (verbose)
-    p += sprintf(p, "  %s", lookup_device_full(d->vendid, d->devid));
+    p += sprintf(p, "  %s",
+		 pci_lookup_name(pacc, namebuf, sizeof(namebuf),
+				 PCI_LOOKUP_VENDOR | PCI_LOOKUP_DEVICE,
+				 q->vendor_id, q->device_id, 0, 0));
   print_it(line, p);
 }
 
@@ -957,6 +1192,176 @@ show_forest(void)
   show_tree_bridge(&host_bridge, line, line);
 }
 
+/* Bus mapping mode */
+
+struct bus_bridge {
+  struct bus_bridge *next;
+  byte this, dev, func, first, last, bug;
+};
+
+struct bus_info {
+  byte exists;
+  byte guestbook;
+  struct bus_bridge *bridges, *via;
+};
+
+static struct bus_info *bus_info;
+
+static void
+map_bridge(struct bus_info *bi, struct device *d, int np, int ns, int nl)
+{
+  struct bus_bridge *b = xmalloc(sizeof(struct bus_bridge));
+  struct pci_dev *p = d->dev;
+
+  b->next = bi->bridges;
+  bi->bridges = b;
+  b->this = get_conf_byte(d, np);
+  b->dev = p->dev;
+  b->func = p->func;
+  b->first = get_conf_byte(d, ns);
+  b->last = get_conf_byte(d, nl);
+  printf("## %02x.%02x:%d is a bridge from %02x to %02x-%02x\n",
+	 p->bus, p->dev, p->func, b->this, b->first, b->last);
+  if (b->this != p->bus)
+    printf("!!! Bridge points to invalid primary bus.\n");
+  if (b->first > b->last)
+    {
+      printf("!!! Bridge points to invalid bus range.\n");
+      b->last = b->first;
+    }
+}
+
+static void
+do_map_bus(int bus)
+{
+  int dev, func;
+  int verbose = pacc->debugging;
+  struct bus_info *bi = bus_info + bus;
+  struct device *d;
+
+  if (verbose)
+    printf("Mapping bus %02x\n", bus);
+  for(dev = 0; dev < 32; dev++)
+    if (filter.slot < 0 || filter.slot == dev)
+      {
+	int func_limit = 1;
+	for(func = 0; func < func_limit; func++)
+	  if (filter.func < 0 || filter.func == func)
+	    {
+	      struct pci_dev *p = pci_get_dev(pacc, bus, dev, func);
+	      u16 vendor = pci_read_word(p, PCI_VENDOR_ID);
+	      if (vendor && vendor != 0xffff)
+		{
+		  if (!func && (pci_read_byte(p, PCI_HEADER_TYPE) & 0x80))
+		    func_limit = 8;
+		  if (verbose)
+		    printf("Discovered device %02x:%02x.%d\n", bus, dev, func);
+		  bi->exists = 1;
+		  if (d = scan_device(p))
+		    {
+		      show_device(d);
+		      switch (get_conf_byte(d, PCI_HEADER_TYPE) & 0x7f)
+			{
+			case PCI_HEADER_TYPE_BRIDGE:
+			  map_bridge(bi, d, PCI_PRIMARY_BUS, PCI_SECONDARY_BUS, PCI_SUBORDINATE_BUS);
+			  break;
+			case PCI_HEADER_TYPE_CARDBUS:
+			  map_bridge(bi, d, PCI_CB_PRIMARY_BUS, PCI_CB_CARD_BUS, PCI_CB_SUBORDINATE_BUS);
+			  break;
+			}
+		      free(d);
+		    }
+		  else if (verbose)
+		    printf("But it was filtered out.\n");
+		}
+	      pci_free_dev(p);
+	    }
+      }
+}
+
+static void
+do_map_bridges(int bus, int min, int max)
+{
+  struct bus_info *bi = bus_info + bus;
+  struct bus_bridge *b;
+
+  bi->guestbook = 1;
+  for(b=bi->bridges; b; b=b->next)
+    {
+      if (bus_info[b->first].guestbook)
+	b->bug = 1;
+      else if (b->first < min || b->last > max)
+	b->bug = 2;
+      else
+	{
+	  bus_info[b->first].via = b;
+	  do_map_bridges(b->first, b->first, b->last);
+	}
+    }
+}
+
+static void
+map_bridges(void)
+{
+  int i;
+
+  printf("\nSummary of buses:\n\n");
+  for(i=0; i<256; i++)
+    if (bus_info[i].exists && !bus_info[i].guestbook)
+      do_map_bridges(i, 0, 255);
+  for(i=0; i<256; i++)
+    {
+      struct bus_info *bi = bus_info + i;
+      struct bus_bridge *b = bi->via;
+
+      if (bi->exists)
+	{
+	  printf("%02x: ", i);
+	  if (b)
+	    printf("Entered via %02x:%02x.%d\n", b->this, b->dev, b->func);
+	  else if (!i)
+	    printf("Primary host bus\n");
+	  else
+	    printf("Secondary host bus (?)\n");
+	}
+      for(b=bi->bridges; b; b=b->next)
+	{
+	  printf("\t%02x.%d Bridge to %02x-%02x", b->dev, b->func, b->first, b->last);
+	  switch (b->bug)
+	    {
+	    case 1:
+	      printf(" <overlap bug>");
+	      break;
+	    case 2:
+	      printf(" <crossing bug>");
+	      break;
+	    }
+	  putchar('\n');
+	}
+    }
+}
+
+static void
+map_the_bus(void)
+{
+  if (pacc->method == PCI_ACCESS_PROC_BUS_PCI ||
+      pacc->method == PCI_ACCESS_DUMP)
+    printf("WARNING: Bus mapping can be reliable only with direct hardware access enabled.\n\n");
+  else if (!check_root())
+    die("Only root can map the bus.");
+  bus_info = xmalloc(sizeof(struct bus_info) * 256);
+  bzero(bus_info, sizeof(struct bus_info) * 256);
+  if (filter.bus >= 0)
+    do_map_bus(filter.bus);
+  else
+    {
+      int bus;
+      for(bus=0; bus<256; bus++)
+	do_map_bus(bus);
+    }
+  map_bridges();
+}
+
 /* Main */
 
 int
@@ -970,32 +1375,31 @@ main(int argc, char **argv)
       puts("lspci version " PCIUTILS_VERSION);
       return 0;
     }
-  filter_init(&filter);
+
+  pacc = pci_alloc();
+  pacc->error = die;
+  pci_filter_init(pacc, &filter);
+
   while ((i = getopt(argc, argv, options)) != -1)
     switch (i)
       {
       case 'n':
-	show_numeric_ids = 1;
+	pacc->numeric_ids = 1;
 	break;
       case 'v':
 	verbose++;
 	break;
       case 'b':
+	pacc->buscentric = 1;
 	buscentric_view = 1;
 	break;
       case 's':
-	if (msg = filter_parse_slot(&filter, optarg))
-	  {
-	    fprintf(stderr, "lspci: -f: %s\n", msg);
-	    return 1;
-	  }
+	if (msg = pci_filter_parse_slot(&filter, optarg))
+	  die("-f: %s", msg);
 	break;
       case 'd':
-	if (msg = filter_parse_id(&filter, optarg))
-	  {
-	    fprintf(stderr, "lspci: -d: %s\n", msg);
-	    return 1;
-	  }
+	if (msg = pci_filter_parse_id(&filter, optarg))
+	  die("-d: %s", msg);
 	break;
       case 'x':
 	show_hex++;
@@ -1004,28 +1408,37 @@ main(int argc, char **argv)
 	show_tree++;
 	break;
       case 'i':
-	pci_ids = optarg;
-	break;
-      case 'p':
-	pci_dir = optarg;
+	pacc->id_file_name = optarg;
 	break;
       case 'm':
 	machine_readable++;
 	break;
+      case 'M':
+	map_mode++;
+	break;
       default:
+	if (parse_generic_option(i, pacc, optarg))
+	  break;
       bad:
-	fprintf(stderr, help_msg);
+	fprintf(stderr, help_msg, pacc->id_file_name);
 	return 1;
       }
   if (optind < argc)
     goto bad;
 
-  scan_proc();
-  sort_them();
-  if (show_tree)
-    show_forest();
+  pci_init(pacc);
+  if (map_mode)
+    map_the_bus();
   else
-    show();
+    {
+      scan_devices();
+      sort_them();
+      if (show_tree)
+	show_forest();
+      else
+	show();
+    }
+  pci_cleanup(pacc);
 
   return 0;
 }
